use std::collections::HashSet;
use std::fmt::{Display, Formatter};
use std::str::FromStr;


#[derive(Copy, Clone, Debug)]
enum Direction {
    Left,
    Right,
    Up,
    Down,
}

impl FromStr for Direction {
    type Err = String;

    fn from_str(s: &str) -> Result<Self, String> {
        match s {
            "L" => Ok(Direction::Left),
            "R" => Ok(Direction::Right),
            "U" => Ok(Direction::Up),
            "D" => Ok(Direction::Down),
            _ => Err(format!("Not a valid direction: {s}")),
        }
    }
}


#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash)]
struct Pos {
    x: isize,
    y: isize,
}


#[derive(Debug)]
struct Rope<const L: usize> {
    knots: [Pos; L],
    tail_visited: HashSet<Pos>,
}

impl<const L: usize> Rope<L> {
    fn new() -> Self {
        Rope {
            knots: [Pos {x: 0, y: 0}; L],
            tail_visited: HashSet::new()
        }
    }

    fn do_move(&mut self, dir: Direction, count: isize) {
        for _ in 0..count {
            let head = &mut self.knots[0];
            match dir {
                Direction::Left => head.x -= 1,
                Direction::Right => head.x += 1,
                Direction::Up => head.y += 1,
                Direction::Down => head.y -= 1,
            }

            for i in 1..(self.knots.len()) {
                self.move_follower(i);                
            }

            self.tail_visited.insert(*self.knots.last().unwrap());
        }
    }

    fn move_follower(&mut self, i: usize) {
        let leader = self.knots[i - 1];
        let follower = &mut self.knots[i];

        let dx = leader.x - follower.x;
        let dy = leader.y - follower.y;

        // check to make sure that we didn't do anything wrong previously
        if dx.abs() + dy.abs() > 4
            || dx.abs() == 0 && dy.abs() > 2
            || dx.abs() > 2 && dy.abs() == 0 
        {
            panic!("Invalid head and tail positions: {dx}, {dy}");
        }

        // head and tail are in line vertically
        if dx == 0 && dy.abs() == 2 {
            follower.y += dy / 2;  // use integer division to keep the sign of dy
        }
        // in line horizontally
        if dy == 0 && dx.abs() == 2 {
            follower.x += dx / 2; // similarly
        }

        // head and tail are offset by 1 horizontaly and 2 vertically
        if dx.abs() == 1 && dy.abs() == 2 {
            follower.x += dx;
            follower.y += dy / 2;
        }
        // offset by 1 vertically and 2 horizontally
        if dy.abs() == 1 && dx.abs() == 2 {
            follower.y += dy;
            follower.x += dx / 2;
        }

        // offset by 2 in both dimensions
        if dx.abs() == 2 && dy.abs() == 2 {
            follower.x += dx / 2;
            follower.y += dy / 2;
        }
    }
}


use std::convert::TryFrom;

impl<const L: usize> Display for Rope<L> {
    fn fmt(&self, f: &mut Formatter) -> std::fmt::Result {
        let min_x = self.knots.iter().map(|pos| pos.x).min().unwrap();
        let max_x = self.knots.iter().map(|pos| pos.x).max().unwrap();
        let min_y = self.knots.iter().map(|pos| pos.y).min().unwrap();
        let max_y = self.knots.iter().map(|pos| pos.y).max().unwrap();

        let n_rows = usize::try_from(max_y - min_y + 1).unwrap();
        let n_cols = usize::try_from(max_x - min_x + 1).unwrap();
        let mut grid = vec![vec!['.'; n_cols]; n_rows];

        for (i, kt) in self.knots.iter().enumerate() {
            let rel_x = usize::try_from(kt.x - min_x).unwrap();
            let rel_y = usize::try_from(max_y - kt.y).unwrap();
            let tgt = &mut grid[rel_y][rel_x];
            if *tgt == '.' {
                *tgt = char::from_digit(i as u32, 10).unwrap();
            }
        }

        for row in grid {
            write!(f, "{}\n", row.iter().collect::<String>())?;
        }

        Ok(())
    }
}


fn main() -> Result<(), String> {
    let instructions = DATA.trim()
        .lines()
        .map(|line| {
            let mut splat = line.split(' ');
            match (splat.next(), splat.next()) {
                (Some(dir), Some(dist)) => {
                    let dist = dist.parse().map_err(|_e| format!("Not an integer: {dist}"));
                    Ok((dir.parse()?, dist?))
                },
                _ => Err(format!("Could not parse move instruction from line: {line}"))
            }
        })
        .collect::<Result<Vec<(Direction, isize)>, String>>()?;

    let mut rope1 = Rope::<2>::new();
    for &(dir, dist) in &instructions {
        rope1.do_move(dir, dist);
    }
    println!("Part 1: {}", rope1.tail_visited.len());

    let mut rope2 = Rope::<10>::new();
    for &(dir, dist) in &instructions {
        rope2.do_move(dir, dist);
    }
    println!("Part 2: {}", rope2.tail_visited.len());

    Ok(())
}


const DATA: &'static str = include_str!("../data/day9.txt");
// const DATA: &'static str = "
// R 5
// U 8
// L 8
// D 3
// R 17
// D 10
// L 25
// U 20
// ";
